package itcr.mathexplorer.expresion;

import java.util.Vector;
import itcr.mathexplorer.exceptions.*;
import itcr.mathexplorer.herramientas.Opciones;

/**
 *@(#)Nodo.java
 *
 *Esta clase define cada uno de los nodos del árbol
 *que conforma una expresión.
 *
 *@author Alexander Borbón Alpízar
 */
public class Nodo {
	
	//VARIABLES PRIVADAS
	
	private Vector<Nodo> operandos;
	private String tipo;
	
	//CONSTRUCTORES
	
	/**
	 *Constructor sin argumentos que deja el tipo vacío y 
	 *deja en cero el número de operandos.
	 */
    public Nodo() {
    	tipo=new String();
    	operandos=null;
    }
    
	/**
	 *Constructor que pone el tipo dado y 
	 *deja en cero el número de operandos.
	 *@param tipo El tipo de nodo a construir.
	 */
	public Nodo(String tipo){
    	this.tipo=tipo;
    	operandos=null;
    }
    
	/**
	 *Constructor que pone el tipo dado y 
	 *el número de operandos indicados.
	 *@param tipo El tipo de nodo a construir.
	 *@param numOperandos El número de operandos que llevará el nodo.
	 */
    public Nodo(String tipo, int numOperandos){
    	this.tipo=tipo;
    	operandos=new Vector<Nodo>(numOperandos);
    }
    
	/**
	 *Constructor que pone el tipo dado y 
	 *toma los operandos de los nodos dados en el arreglo.
	 *@param tipo El tipo de nodo a construir.
	 *@param operandos Los operandos relacionados con el tipo dado.
	 */
    public Nodo(String tipo, Vector<Nodo> operandos){
    	this.tipo=tipo;
    	this.operandos=operandos;
    }
    
    //FUNCIONES PUBLICAS
    
    /**
     *Función que le agrega un operando(un nodo) al tipo actual(al nodo actual).
     *@param operando El operando para ser agregado.
     *@return Este nodo para encadenamiento.
     */
    public Nodo agregaOperando(Nodo operando){
    	if (operandos==null){
    		operandos=new Vector<Nodo>();
    		operandos.addElement(operando);
    	}else{
    		operandos.addElement(operando);
    	}
    	
    	return this;
    }
    
    /**
     *Función que le agrega un operando(un nodo) al tipo actual(al nodo actual) en la posición dada.
     *@param operando El operando para ser agregado.
     *@param pos La posición en la cual se quiere agregar el operando.
     *@return Este nodo para encadenamiento.
     *@exception ArrayIndexOutOfBoundsException Si pos está fuera del tamaño del Vector.
     */
    public Nodo agregaOperando(Nodo operando, int pos) throws ArrayIndexOutOfBoundsException{
    	try{
    		operandos.add(pos-1,operando);
    		
    		return this;
    	}catch(ArrayIndexOutOfBoundsException e){
    		e.printStackTrace();
    		throw e;
    	}
    }
    
    /**
     *Función que devuelve el operando en la posición pos.
     *@param pos La posición del operando que se quiere obtener.
     *@return Este nodo para encadenamiento.
     *@exception NotFoundElementException Si no encuentra elemento en la posición buscada.
     */
    public Nodo getOperando(int pos) throws NotFoundElementException{
    	try{
    		return (Nodo)operandos.elementAt(pos-1);
    	}catch(ArrayIndexOutOfBoundsException e){
    		e.printStackTrace();
    		throw new NotFoundElementException("Elemento no encontrado");
    	}
    }
    
    /**
     *Función que elimina el operando en la posición pos.
     *@param pos La posición del operando que se quiere eliminar.
     *@return Este nodo para encadenamiento.
     *@exception NotFoundElementException Si no encuentra elemento a borrar en la posición buscada.
     */
    public Nodo eliminaOperando(int pos) throws NotFoundElementException{
    	try{
    		operandos.removeElementAt(pos-1);
    		
    		return this;
    	}catch(ArrayIndexOutOfBoundsException e){
    		e.printStackTrace();
    		throw new NotFoundElementException();
    	}
    }
    
    /**
     *Función que intercambia el operando en la posición A con el operando en la posición B.
     *@param posA La posición del primer elemento a intercambiar.
     *@param posB La posición del segundo elemento a intercambiar.
     *@return Este nodo para encadenamiento.
     *@exception NotFoundElementException Si no encuentra los elementos en las posiciones buscadas.
     */
    public Nodo intercambiaOperandos(int posA, int posB) throws NotFoundElementException{
    	Nodo comodin;
    	
    	try{
    		if(posA<posB){
	    		comodin=this.getOperando(posB);
	    		this.eliminaOperando(posB);
	    		this.agregaOperando(this.getOperando(posA),posB);
	    		this.eliminaOperando(posA);
	    		this.agregaOperando(comodin,posA);
    		}else{
	    		comodin=this.getOperando(posA);
	    		this.eliminaOperando(posA);
	    		this.agregaOperando(this.getOperando(posB),posA);
	    		this.eliminaOperando(posB);
	    		this.agregaOperando(comodin,posB);
    		}
    		
    		return this;
    	}catch(ArrayIndexOutOfBoundsException e){
    		e.printStackTrace();
    		throw new NotFoundElementException();
    	}
    }
    
    /**
     *Devuelve el número de operandos que tiene el nodo.
     *@return El número de operandos que están actualmente en este nodo.
     */
    public int getNumeroOperandos(){
    	if (operandos==null){
    		return 0;
    	}else{
    		return operandos.size();
    	}
    }
    
    /**
     *Devuelve el tipo del nodo que es (la cabecera del nodo).
     *@return El tipo del nodo.
     */
    public String getTipo(){
   		return tipo;
    }
    
    /**
     *Cambia el tipo del nodo que es (la cabecera del nodo).
     *@param nuevoTipo El tipo nuevo del nodo (la nueva cabecera).
     *@return El mismo nodo para recursividad.
     */
    public Nodo cambiaTipo(String nuevoTipo){
    	tipo=nuevoTipo;
    	
   		return this;
    }
    
	/**
     *Devuelve un clon del nodo con todos sus subnodos correspondientes (un clon del árbol de la expresión).
     *@return Un nodo con el clon de la expresión.
     */
    public Nodo clonarNodo(){
    	Nodo comodin;
    	
    	if (this.getNumeroOperandos()==0){
    		return(new Nodo(this.getTipo()));
    	}else{
    		comodin=new Nodo(this.getTipo(),this.getNumeroOperandos());
    		for(int i=1;i<=this.getNumeroOperandos();i++){
    			comodin.agregaOperando(this.getOperando(i).clonarNodo());
    		}
    		return comodin;
    	}
    }
    
    /**
     *Devuelve si el nodo es igual al nodo nado.
     *@param comparar El nodo con el cual se hará la comparación.
     *@return True si el nodo es igual al nodo con el que se compara y false en caso contrario.
     */
    public boolean esIgual(Nodo comparar){
    	String tipo=this.getTipo(), tipo2;
    	int pos1, pos2;
    	int num1, num2;
    	
   		if(!tipo.equals(comparar.getTipo())||this.getNumeroOperandos()!=comparar.getNumeroOperandos()){
   			return false;
   		}
   		
   		if(tipo.length()>5 && "DecimalLargo".indexOf(tipo)!=-1){
   			tipo=this.getOperando(1).getTipo();
   			tipo2=comparar.getOperando(1).getTipo();
   			
			pos1=tipo.indexOf(".");
			pos2=tipo2.indexOf(".");
			num1=pos1+Opciones.getNumDecimales()-9;
			num2=pos2+Opciones.getNumDecimales()-9;
			if(pos1!=-1 && num1<=tipo.length()){
				tipo=tipo.substring(0,num1);
			}
			
			if(pos2!=-1 && num2<=tipo2.length()){
				tipo2=tipo2.substring(0,num2);
			}
   			
   			if(tipo.equals(tipo2)){
   				return true;
   			}else{
   				return false;
   			}
   		}
   		
   		for(int i=1;i<=this.getNumeroOperandos();i++){
   			if(!this.getOperando(i).esIgual(comparar.getOperando(i))){
   				return false;
   			}
   		}
   		return true;
    }
    
    /**
     *Devuelve un String que representa al nodo.
     *@return El String que representa el nodo.
     */
    public String toString(){
    	String expresion="";
    	
    	for(int i=1; i<=this.getNumeroOperandos();i++){
    		expresion=expresion.concat(this.getOperando(i).toString());
    		if(i!=this.getNumeroOperandos()){
    			expresion=expresion.concat(", ");
    		}
    	}
    	
    	if(expresion.equals("")){
    		return this.getTipo();
    	}else{
    		expresion=this.getTipo().concat("(").concat(expresion).concat(")");
    		return expresion;
    	}
    }
}//Nodo