package itcr.mathexplorer.funciones;

import itcr.mathexplorer.expresion.*;
import itcr.mathexplorer.tablas.*;
import itcr.mathexplorer.operadores.*;
import itcr.mathexplorer.constantes.*;
import itcr.mathexplorer.tipos.*;
import itcr.mathexplorer.herramientas.Opciones;
import java.math.BigDecimal;
import java.math.MathContext;

/**
 *@(#)SimplificaOperacionNumeros.java
 *
 *Esta es la implementación de la función que realiza 
 *la simplificación de una operación en modo exacto con números.
 *
 *Este nodo está pensado para utilizarse en conjunto con expresion
 *y los traductores del paquete.
 *
 *@author Alexander Borbón Alpízar
 */
public class SimplificarNumeros extends Funcion{
	//VARIABLES PRIVADAS
	
	private TablaOperadores operadores;
	private TablaFunciones funciones;
	private TablaConstantes constantes;
	private TablaFuncionesUsuario funcionesUsuario;
	
	/**
	 * Define la expresión que representa el UNO.
	 */
	static public Nodo UNO;
	
	/**
	 * Define la expresión que representa el MENOSUNO. 
	 */
	static public Nodo MENOSUNO;
	
	/**
	 * Define la expresión que representa el CERO. 
	 */
	static public Nodo CERO;
	
	/**
	 * Define la expresión que representa el DIEZ. 
	 */
	static public Nodo DIEZ;
	
	/**
	 * Define la expresión que representa el PI. 
	 */
	static public Nodo PI;
	
	/**
	 * Define la expresión que representa el TRUE. 
	 */
	static public Nodo TRUE;
	
	/**
	 * Define la expresión que representa el FALSE. 
	 */
	static public Nodo FALSE;
	
	
	/**
	 *Constructor que recibe las tablas de operadores, funciones y constantes para comparar.
	 *@param tablaoperadores La tabla de operadores.
	 *@param tablafunciones La tabla de funciones.
	 *@param tablaconstantes La tabla de constantes.
	 *@param tablafuncionesusuario La tabla de funciones definidas por el usuario.
	 */
	public SimplificarNumeros(TablaOperadores tablaoperadores, TablaFunciones tablafunciones, TablaConstantes tablaconstantes, TablaFuncionesUsuario tablafuncionesusuario) {
		operadores=tablaoperadores;
		funciones=tablafunciones;
		constantes=tablaconstantes;
		funcionesUsuario=tablafuncionesusuario;
		
		if(Opciones.getPresicion()){
			MENOSUNO=new Nodo("Entero",1);
			UNO=new Nodo("Entero",1);
			CERO=new Nodo("Entero",1);
			DIEZ=new Nodo("Entero",1);
		}else{
			MENOSUNO=new Nodo("Decimal",1);
			UNO=new Nodo("Decimal",1);
			CERO=new Nodo("Decimal",1);
			DIEZ=new Nodo("Decimal",1);
		}
		MENOSUNO.agregaOperando(new Nodo("-1"));
		UNO.agregaOperando(new Nodo("1"));
		CERO.agregaOperando(new Nodo("0"));
		DIEZ.agregaOperando(new Nodo("10"));
		
		TRUE=new Nodo("Boolean");
		TRUE.agregaOperando(new Nodo("True"));
		FALSE=new Nodo("Boolean");
		FALSE.agregaOperando(new Nodo("False"));
	}
	
	//FUNCIONES PUBLICAS
	
	/**
	 *Simplificar sólo admite un parámetro
	 */
	public boolean permiteNumeroParametros(int numeroParametros){
		return numeroParametros==1;
	}
	
	/**
	 *Función que simplifica la expresión recibida.
	 *@param expresionPrefijo La expresión a simplificar.
	 *@param funcionSimplificar La función que se encarga de simplificar. Se necesita para la recursividad.
	 *@param funcionSimplificarNumeros La función que se encarga de simplificar operaciones numéricas. Se necesita para la recursividad.
	 *@return El resultado de la operación con números.
	 *@exception ArithmeticException Se lanza si hay error al simplificar en la expresión.
	 */
	public Nodo simplificarNumeros(Nodo expresionPrefijo, Simplificar funcionSimplificar, SimplificarNumeros funcionSimplificarNumeros) throws ArithmeticException{
		try{
			if(expresionPrefijo.getOperando(1).getTipo().equals("Indefinido") || expresionPrefijo.getOperando(1).getTipo().equals("Boolean")){
				return new Nodo("Indefinido");
			}
			return expresionPrefijo;
		}catch(Exception e){
			e.printStackTrace();
			throw new ArithmeticException("Error al simplificar una simplificación de números");
		}	
	}
	
	/**
	 *Función que simplifica la expresión recibida. Si esta función se evalúa aquí quiere decir que no encontró el operador
	 *donde se tenía que evaluar, lo que es un error.
	 *@param expresionPrefijo La expresión a simplificar.
	 *@param funcionSimplificar La función que se encarga de simplificar. Se necesita para la recursividad.
	 *@param funcionSimplificarNumeros La función que se encarga de simplificar operaciones numéricas. Se necesita para la recursividad.
	 *@return La expresión simplificada.
	 *@exception ArithmeticException Se lanza si hay error al simplificar en la expresión.
	 */
	public Nodo simplificar(Nodo expresionPrefijo, Simplificar funcionSimplificar, SimplificarNumeros funcionSimplificarNumeros) throws ArithmeticException{
		try{
			if(expresionPrefijo.getOperando(1).getTipo().equals("Indefinido") || expresionPrefijo.getOperando(1).getTipo().equals("Boolean")){
				return new Nodo("Indefinido");
			}
			return expresionPrefijo;
		}catch(Exception e){
			e.printStackTrace();
			throw new ArithmeticException("Error al simplificar una simplificación de números");
		}
	}

	/**
	 * Función que realiza la simplificación de la operación con números.
	 * @param expresionPrefijo La expresión a simplificar.
	 * @param funcionSimplificar La función que se encarga de simplificar. Se necesita para la recursividad.
	 * @return Un Nodo con la expresión de números simplificada
	 * @exception ArithmeticException Se lanza si hay error en la expresión.
	 */
	public Nodo simplificaNumeros(Nodo expresionPrefijo, Simplificar funcionSimplificar) throws ArithmeticException{
		String tipo=expresionPrefijo.getTipo();
		Nodo comodin;
		String numero="";
		BigDecimal nuevoDecimal;
		int pos;
		
		try{
			if(tipo.equals("Indefinido")){
				return new Nodo("Indefinido");
			}
			
			if(Opciones.getPresicion() && tipo.length()>5 && "DecimalLargo".indexOf(tipo)!=-1){
				numero=expresionPrefijo.getOperando(1).getTipo();
				pos=numero.indexOf(".");
				int pos1=numero.indexOf("E");
				
				if(pos!=-1){
					String numerador=numero.substring(0,pos).concat(numero.substring(pos+1,numero.length()));
					String denominador=new String("1");
					for(int i=pos;i<numero.length()-1;i++){
						denominador=denominador.concat("0");
					}
					BigRational nuevoRacional=new BigRational(numerador,denominador);
					comodin=new Nodo("Racional",2);
					comodin.agregaOperando(new Nodo(nuevoRacional.getNumerator().toString()));
					comodin.agregaOperando(new Nodo(nuevoRacional.getDenominator().toString()));
					
					if(nuevoRacional.getDenominator().toString().equals("1")){
						numero=nuevoRacional.getNumerator().toString();
						
						if(numero.length()>18){
							comodin=new Nodo("EnteroLargo",1);
						}else{
							comodin=new Nodo("Entero",1);
						}
						
						comodin.agregaOperando(new Nodo(numero));
					}
				}else if(pos1!=-1){
					String numerador=numero.substring(0,pos1);
					String denominador=new String("1");
					
					int num=new Integer(numero.substring(pos1+2,numero.length()));
					for(int i=pos1;i<=num;i++){
						denominador=denominador.concat("0");
					}
					BigRational nuevoRacional=new BigRational(numerador,denominador);
					comodin=new Nodo("Racional",2);
					comodin.agregaOperando(new Nodo(nuevoRacional.getNumerator().toString()));
					comodin.agregaOperando(new Nodo(nuevoRacional.getDenominator().toString()));
					
					if(nuevoRacional.getDenominator().toString().equals("1")){
						numero=nuevoRacional.getNumerator().toString();
						
						if(numero.length()>18){
							comodin=new Nodo("EnteroLargo",1);
						}else{
							comodin=new Nodo("Entero",1);
						}
						
						comodin.agregaOperando(new Nodo(numero));
					}
				}else{
					if(numero.length()>18){
						comodin=new Nodo("EnteroLargo",1);
					}else{
						comodin=new Nodo("Entero",1);
					}
					
					comodin.agregaOperando(new Nodo(numero));
				}
				
				return comodin;
			}else if(!Opciones.getPresicion()){
				if(tipo.equals("Constante")){
					return ((Constante)constantes.devuelveConstante(expresionPrefijo.getOperando(1).getTipo())).simplificarNumeros(expresionPrefijo, funcionSimplificar, this);
				}else if(tipo.equals("Racional")){//("DecimalLargo Racional").indexOf(tipo)!=-1){
					String numerador, denominador;
					
					numerador=expresionPrefijo.getOperando(1).getTipo();
					denominador=expresionPrefijo.getOperando(2).getTipo();
					
					BigRational nuevoRacional=new BigRational(numerador,denominador);
					nuevoDecimal=nuevoRacional.toBigDecimal(Math.max(numerador.length(),denominador.length())+Opciones.getNumDecimales());
					nuevoDecimal=nuevoDecimal.round(new MathContext(Opciones.getNumDecimales()+nuevoDecimal.toString().indexOf(".")+1));
					
					numero=nuevoDecimal.toString();
					if(numero.length()>18){
						comodin=new Nodo("DecimalLargo");
					}else{
						comodin=new Nodo("Decimal");
					}

					comodin.agregaOperando(new Nodo(numero));
					return comodin;
				}else if(tipo.length()>5 && "EnteroLargo".indexOf(tipo)!=-1){
					if(tipo.equals("EnteroLargo")){
						comodin=new Nodo("DecimalLargo");
					}else{
						comodin=new Nodo("Decimal");
					}

					comodin.agregaOperando(new Nodo(expresionPrefijo.getOperando(1).getTipo()));
					return comodin;
				}else if(tipo.length()>5 && "DecimalLargo".indexOf(tipo)!=-1){
					numero=expresionPrefijo.getOperando(1).getTipo();
					
					pos=numero.indexOf(".");
					
					if(pos!=-1){
						nuevoDecimal=new BigDecimal(numero);
						nuevoDecimal=nuevoDecimal.round(new MathContext(Opciones.getNumDecimales()+pos+1));
						
						numero=nuevoDecimal.toPlainString();
						
						while(numero.charAt(numero.length()-1)=='0'){
							numero=numero.substring(0,numero.length()-1);
						}
						if(numero.charAt(numero.length()-1)=='.'){
							numero=numero.substring(0,numero.length()-1);
						}
					}
					
					pos=numero.indexOf("E");
					
					if(pos!=-1){
						nuevoDecimal=new BigDecimal(numero);
						nuevoDecimal=nuevoDecimal.add(new BigDecimal(1));
						nuevoDecimal=nuevoDecimal.round(new MathContext(Opciones.getNumDecimales()+pos+1));
						
						numero=nuevoDecimal.toPlainString();
						
						while(numero.charAt(numero.length()-1)=='0'){
							numero=numero.substring(0,numero.length()-1);
						}
						if(numero.charAt(numero.length()-1)=='.'){
							numero=numero.substring(0,numero.length()-1);
						}
						
						nuevoDecimal=new BigDecimal(numero);
						nuevoDecimal=nuevoDecimal.add(new BigDecimal(-1));
						
						numero=nuevoDecimal.toPlainString();
					}
					
					if(numero.length()>18){
						comodin=new Nodo("DecimalLargo");
					}else{
						comodin=new Nodo("Decimal");
					}
					
					comodin.agregaOperando(new Nodo(numero));

					return comodin;
				}
				
			}
			
			if(tipo.equals("Racional")){
				if(expresionPrefijo.getOperando(2).getTipo().equals("0")){
					return new Nodo("Indefinido");
				}else if(expresionPrefijo.getOperando(2).getTipo().equals("1")){
					numero=expresionPrefijo.getOperando(1).getTipo();
					
					if(numero.length()>18){
						comodin=new Nodo("EnteroLargo",1);
					}else{
						comodin=new Nodo("Entero",1);
					}
					
					comodin.agregaOperando(new Nodo(numero));
					
					return comodin;
				}else{
					return expresionPrefijo;
				}
			}else if(operadores.apareceLlave(tipo)){ //Si es un operador binario
				return ((Operador)operadores.devuelveOperador(tipo)).simplificarNumeros(expresionPrefijo, funcionSimplificar, this);
			}else if(funciones.apareceLlave(tipo)){ //Si es una función
				return ((Funcion)funciones.devuelveFuncion(tipo)).simplificarNumeros(expresionPrefijo, funcionSimplificar, this);
			}else if(funcionesUsuario.apareceLlave(tipo)){ //Si es una función definida por el usuario
				return simplificaNumeros((Nodo)((FuncionUsuario)funcionesUsuario.devuelveFuncion(tipo)).obtenExpresion(), funcionSimplificar);
			}
			
			return expresionPrefijo;
		}catch(Exception e){
			e.printStackTrace();
			throw new ArithmeticException("Error en la expresión al simplificar");
		}
	}
	
	/**
	 *Función que deriva la expresión recibida.
	 *@param expresionPrefijo La expresión a derivar.
	 *@param variable La variable con respecto a la que se va a derivar.
	 *@param funcionDerivar La función que se encarga de derivar.
	 *@return La expresión derivada.
	 *@exception ArithmeticException Se lanza si hay error al derivar en la expresión.
	 */
	public Nodo derivar(Nodo expresionPrefijo, Nodo variable, Derivar funcionDerivar) throws ArithmeticException{
		try{
			return expresionPrefijo;
		}catch(Exception e){
			e.printStackTrace();
			throw new ArithmeticException("Error al derivar un simplificarNumeros");
		}
	}
	
	/**
	 *Esta función devuelve un nodo representando la expresión pi/n
	 *@param n Entero para crear la expresión pi/n
	 *@return El nodo que representa la expresion pi/n
	 *@exception ArithmeticException Se lanza si hay error al crear la expresión.
	 */
	public Nodo piEntre(int n){
		Nodo comodin, comodin2;
		String numero;
		
		try{
			comodin=new Nodo("Entero",1);
			comodin.agregaOperando(new Nodo("1"));
			
			numero=(new Integer(n)).toString();
			
			comodin2=new Nodo("Racional",2);
			comodin2.agregaOperando(new Nodo("1"));
			comodin2.agregaOperando(new Nodo(numero));
			
			comodin=new Nodo("·",2);
			comodin.agregaOperando(comodin2);
			
			comodin2=new Nodo("Constante",1);
			comodin2.agregaOperando(new Nodo("#pi"));
			
			comodin.agregaOperando(comodin2);
			
			return comodin;
		}catch(Exception e){
			e.printStackTrace();
			throw new ArithmeticException("Error al construir pi/n");
		}
	}
	
	/**
	 *Esta función devuelve un nodo representando la expresión k*pi/n
	 *@param k Entero para crear la expresión k*pi/n
	 *@param n Entero para crear la expresión k*pi/n
	 *@return El nodo que representa la expresion k*pi/n
	 *@exception ArithmeticException Se lanza si hay error al crear la expresión.
	 */
	public Nodo piPorEntre(int k, int n){
		Nodo comodin, comodin2;
		String numero1, numero2;
		
		try{
			comodin=new Nodo("Entero",1);
			comodin.agregaOperando(new Nodo("1"));
			
			numero1=(new Integer(k)).toString();
			numero2=(new Integer(n)).toString();
			
			comodin2=new Nodo("Racional",2);
			comodin2.agregaOperando(new Nodo(numero1));
			comodin2.agregaOperando(new Nodo(numero2));
			
			comodin=new Nodo("·",2);
			comodin.agregaOperando(comodin2);
			
			comodin2=new Nodo("Constante",1);
			comodin2.agregaOperando(new Nodo("#pi"));
			
			comodin.agregaOperando(comodin2);
			
			return comodin;
		}catch(Exception e){
			e.printStackTrace();
			throw new ArithmeticException("Error al construir pi/n");
		}
	}
	
	/**
	 *Esta función devuelve un nodo representando la expresión n*pi
	 *@param n Entero para crear la expresión n*pi
	 *@return El nodo que representa la expresion n*pi
	 *@exception ArithmeticException Se lanza si hay error al crear la expresión.
	 */
	public Nodo piPor(int n){
		Nodo comodin, comodin2;
		String numero;
		
		try{
			numero=(new Integer(n)).toString();
			
			comodin=new Nodo("Entero",1);
			comodin.agregaOperando(new Nodo(numero));
			
			comodin2=new Nodo("·",2);
			comodin2.agregaOperando(comodin);
			
			comodin=new Nodo("Constante",1);
			comodin.agregaOperando(new Nodo("#pi"));
			
			comodin2.agregaOperando(comodin);
			
			return comodin2;
		}catch(Exception e){
			e.printStackTrace();
			throw new ArithmeticException("Error al construir n*pi");
		}
	}
	
	/**
	 *Esta función devuelve un nodo representando la expresión pi
	 *@return El nodo que representa la expresion pi
	 *@exception ArithmeticException Se lanza si hay error al crear la expresión.
	 */
	public Nodo pi(){
		Nodo comodin;
		
		try{
			comodin=new Nodo("Constante",1);
			comodin.agregaOperando(new Nodo("#pi"));
			
			return comodin;
		}catch(Exception e){
			e.printStackTrace();
			throw new ArithmeticException("Error al construir pi");
		}
	}
	
	/**
	 *Función que indica si la expresión recibida es un número entero.
	 *@param expresionPrefijo La expresión a evaluar.
	 *@return True si es un número entero, false en caso contrario.
	 */
	public boolean esEntero(Nodo expresionPrefijo){
		String tipo;
		
		tipo=expresionPrefijo.getTipo();
		
		if(tipo.length()>5){
			if("EnteroLargo".indexOf(tipo)!=-1){
				return true;
			}else if("DecimalLargo".indexOf(tipo)!=-1){
				if(expresionPrefijo.getOperando(1).getTipo().indexOf(".")==-1){
					return true;
				}else{
					return false;
				}
			}else{
				return false;
			}
		}
		
		return false;
	}
}
