package itcr.mathexplorer.funciones;

import itcr.mathexplorer.expresion.*;
import itcr.mathexplorer.tipos.*;
import itcr.mathexplorer.herramientas.Opciones;
import itcr.mathexplorer.herramientas.MetodosNumericos;
import java.math.BigDecimal;

/**
 *@(#)Sqrt.java
 *
 *Esta es la implementación de la función raíz cuadrada
 *
 *@author Alexander Borbón Alpízar
 */
public class Sqrt extends Funcion{
	
	//CONSTRUCTOR
	/**
	 * Constructor vacío para la función Sqrt.
	 */
	public Sqrt() {	}

	//FUNCIONES PUBLICAS
	
	/**
	 *La raíz sólo admite un parámetro.
	 */
	public boolean permiteNumeroParametros(int numeroParametros){
		return numeroParametros==1;
	}
	
	/**
	 *Función que simplifica la expresión recibida.
	 *@param expresionPrefijo La expresión a simplificar.
	 *@param funcionSimplificar La función que se encarga de simplificar. Se necesita para la recursividad.
	 *@param funcionSimplificarNumeros La función que se encarga de simplificar operaciones numéricas. Se necesita para la recursividad.
	 *@return El resultado de la operación con números.
	 *@exception ArithmeticException Se lanza si hay error al simplificar en la expresión.
	 */
	public Nodo simplificarNumeros(Nodo expresionPrefijo, Simplificar funcionSimplificar, SimplificarNumeros funcionSimplificarNumeros) throws ArithmeticException{
		String tipo;
		//Double numero;
		Nodo comodin;
		
		try{
			tipo=expresionPrefijo.getOperando(1).getTipo();
			
			if(tipo.equals("Indefinido")){
				return new Nodo("Indefinido");
			}else if (tipo.length()>5 && "EnteroLargo DecimalLargo Racional".indexOf(tipo)!=-1){
				if(!Opciones.getPresicion()){
					if(tipo.equals("Racional")){
						String numerador=expresionPrefijo.getOperando(1).getOperando(1).getTipo(), denominador=expresionPrefijo.getOperando(1).getOperando(2).getTipo();
						
						tipo=((new BigRational(numerador,denominador)).toBigDecimal(Math.max(numerador.length(),denominador.length())+Opciones.getNumDecimales())).toString();
					}else{
						tipo=expresionPrefijo.getOperando(1).getOperando(1).getTipo();
					}
					
					if(tipo.charAt(0)=='-'){
						return new Nodo("Indefinido");
					}
					if(Opciones.getNumDecimales() <=Opciones.getLongNumeros() && tipo.length()<=Opciones.getLongNumeros()){
						tipo=(new Double(Math.sqrt((new Double(tipo)).doubleValue()))).toString();
						comodin=new Nodo("Decimal");
					}else{
						tipo = MetodosNumericos.calculeRaiz(new BigDecimal(tipo),Opciones.getNumDecimales()).toString();
						comodin=new Nodo("DecimalLargo");
					}
					
					comodin.agregaOperando(new Nodo(tipo));
					
					return funcionSimplificarNumeros.simplificaNumeros(comodin, funcionSimplificar);
				}
			}
			
			return expresionPrefijo;
		}catch(Exception e){
			e.printStackTrace();
			throw new ArithmeticException("Error al simplificar una raíz");
		}	
	}

	/**
	 *Función que simplifica la expresión recibida. Si esta función se evalúa aquí quiere decir que no encontró el operador
	 *donde se tenía que evaluar, lo que es un error.
	 *@param expresionPrefijo La expresión a simplificar.
	 *@param funcionSimplificar La función que se encarga de simplificar. Se necesita para la recursividad.
	 *@param funcionSimplificarNumeros La función que se encarga de simplificar operaciones numéricas. Se necesita para la recursividad.
	 *@return La expresión simplificada.
	 *@exception ArithmeticException Se lanza si hay error al simplificar en la expresión.
	 */
	public Nodo simplificar(Nodo expresionPrefijo, Simplificar funcionSimplificar, SimplificarNumeros funcionSimplificarNumeros) throws ArithmeticException{
		String tipo;
		
		try{
			if(!expresionPrefijo.getTipo().equals("sqrt")){
				expresionPrefijo.cambiaTipo("sqrt");
			}
			
			tipo=expresionPrefijo.getOperando(1).getTipo();
			
			if(tipo.equals("Indefinido")){
				return new Nodo("Indefinido");
			}
			
			//Si no es un número de manera directa (sino una operación con números) no tiene porque hacerlo
			if (tipo.length()>5 && "EnteroLargo DecimalLargo Racional".indexOf(tipo)!=-1){
				return simplificarNumeros(expresionPrefijo, funcionSimplificar, funcionSimplificarNumeros);
			}
			
			return expresionPrefijo;
		}catch(Exception e){
			e.printStackTrace();
			throw new ArithmeticException("Error al simplificar una raíz");
		}
	}
	
	/**
	 *Función que deriva la expresión recibida.
	 *@param expresionPrefijo La expresión a derivar.
	 *@param variable La variable con respecto a la que se va a derivar.
	 *@param funcionDerivar La función que se encarga de derivar.
	 *@return La expresión derivada.
	 *@exception ArithmeticException Se lanza si hay error al derivar en la expresión.
	 */
	public Nodo derivar(Nodo expresionPrefijo, Nodo variable, Derivar funcionDerivar) throws ArithmeticException{
		Nodo comodin, comodin2, comodin3;
		
		try{
			if (Opciones.getPresicion()){
				comodin=new Nodo("Entero",1);
			}else{
				comodin=new Nodo("Decimal",1);
			}
			comodin.agregaOperando(new Nodo("2"));
			
			comodin2=new Nodo("·",2);
			comodin2.agregaOperando(comodin);
			comodin2.agregaOperando(expresionPrefijo.clonarNodo());
			
			if (Opciones.getPresicion()){
				comodin=new Nodo("Entero",1);
			}else{
				comodin=new Nodo("Decimal",1);
			}
			comodin.agregaOperando(new Nodo("1"));
			
			comodin3=new Nodo("/",2);
			comodin3.agregaOperando(comodin);
			comodin3.agregaOperando(comodin2);
			
			comodin=new Nodo("·",2);
			comodin.agregaOperando(comodin3);
			comodin.agregaOperando(funcionDerivar.deriva(expresionPrefijo.getOperando(1).clonarNodo(), variable));
			
			return comodin;
		}catch(Exception e){
			e.printStackTrace();
			throw new ArithmeticException("Error al derivar una raíz");
		}
	}
}
