package itcr.mathexplorer.herramientas;

import java.math.BigDecimal;

/**
 *@(#)MetodosNumericos.java
 *
 * Clase con los algoritmos para el cálculo con métodos numéricos
 *
 * @author Cristhian Páez Páez y Alexander Borbón Alpízar
 *
 *Métodos principales:
 *
 *BigDecimal calculeSeno(BigDecimal numero, int presicion) Aproxima Seno con la presición deseada.
 *BigDecimal calculeCoseno(BigDecimal numero, int presicion) Aproxima Coseno con la presición deseada.
 *BigDecimal calculeArcoseno(BigDecimal numero, int presicion) Aproxima Arcoseno con la presición deseada.
 *BigDecimal calculeE(BigDecimal numero, int decimales) Aproxima e^x con la presición deseada.
 *BigDecimal calculePi(BigDecimal numero, int presicion) Aproxima pi con la presición deseada.
 */
public class MetodosNumericos {
	
	/**
	 * Aproxima Sen(numero) con la presición deseada.
	 * @param numero El número a evaluar en la función seno.
	 * @param presicion La cantidad de decimales exactos que se desean.
	 * @return Un BigDecimal con el resultado de Sen(numero).
	 */
	static public BigDecimal calculeSeno(BigDecimal numero, int presicion){
		BigDecimal potencia=numero;
		BigDecimal factorial=BigDecimal.ONE;
		BigDecimal resultado=numero;
		BigDecimal cociente;
		
		final BigDecimal cota=BigDecimal.ONE.divide((BigDecimal.TEN).pow(presicion));
		
		int i=1;
		do{
			potencia=potencia.multiply(numero).multiply(numero);
			factorial=factorial.multiply(new BigDecimal(2*i)).multiply(new BigDecimal(2*i+1));
			
			cociente=potencia.divide(factorial, presicion+5, BigDecimal.ROUND_HALF_EVEN);
			i+=1;
			if(i%2==1){
				resultado=resultado.add(cociente);
			}else{
				resultado=resultado.subtract(cociente);
			}
		}while(cociente.compareTo(cota)>0);
		
		return resultado.setScale(presicion, BigDecimal.ROUND_HALF_EVEN);
	}
	
	/**
	 * Aproxima Cos(numero) con la presición deseada.
	 * @param numero El número a evaluar en la función coseno.
	 * @param presicion La cantidad de decimales exactos que se desean.
	 * @return Un BigDecimal con el resultado de Cos(numero).
	 */
	static public BigDecimal calculeCoseno(BigDecimal numero, int presicion){
		BigDecimal potencia=BigDecimal.ONE;
		BigDecimal factorial=BigDecimal.ONE;
		BigDecimal resultado=BigDecimal.ONE;
		BigDecimal cociente;
		
		final BigDecimal cota=BigDecimal.ONE.divide((BigDecimal.TEN).pow(presicion));
		
		int i=1;
		do{
			potencia=potencia.multiply(numero).multiply(numero);
			factorial=factorial.multiply(new BigDecimal(2*i-1)).multiply(new BigDecimal(2*i));
			
			cociente=potencia.divide(factorial, presicion+5, BigDecimal.ROUND_HALF_EVEN);
			i+=1;
			if(i%2==1){
				resultado=resultado.add(cociente);
			}else{
				resultado=resultado.subtract(cociente);
			}
		}while(cociente.compareTo(cota)>0);
		
		return resultado.setScale(presicion, BigDecimal.ROUND_HALF_EVEN);
	}
	
	/**
	 * Aproxima Arcsen(numero) con la presición deseada.
	 * @param numero El número a evaluar en la función arcoseno.
	 * @param presicion La cantidad de decimales exactos que se desean.
	 * @return Un BigDecimal con el resultado de Arcsen(numero).
	 */
	static public BigDecimal calculeArcoseno(BigDecimal numero, int presicion){
		BigDecimal numerador=BigDecimal.ONE;
		BigDecimal denominador=BigDecimal.ONE;
		BigDecimal potencia=numero;
		BigDecimal multiplicaNumerador=new BigDecimal(-1);
		BigDecimal multiplicaDenominador=BigDecimal.ZERO;
		BigDecimal multiplicaDenominador2=BigDecimal.ONE;
		BigDecimal numeroCuadrado=numero.multiply(numero);
		BigDecimal resultado=numero;
		BigDecimal cociente;
		final BigDecimal DOS=new BigDecimal(2);
		final BigDecimal cota=BigDecimal.ONE.divide((BigDecimal.TEN).pow(presicion));
		
		do{
			multiplicaNumerador=multiplicaNumerador.add(DOS);
			multiplicaDenominador=multiplicaDenominador.add(DOS);
			multiplicaDenominador2=multiplicaDenominador2.add(DOS);
			
			numerador=numerador.multiply(multiplicaNumerador);
			denominador=denominador.multiply(multiplicaDenominador);
			potencia=potencia.multiply(numeroCuadrado);
			
			cociente=potencia.multiply(numerador).divide(denominador.multiply(multiplicaDenominador2), presicion+5, BigDecimal.ROUND_HALF_EVEN);

			resultado=resultado.add(cociente);
		}while(cociente.compareTo(cota)>0);
		
		return resultado.setScale(presicion, BigDecimal.ROUND_HALF_EVEN);
	}
	
	/**
	 * Función que calcula de manera numérica la raíz cuadrada de un
	 * número con la cantidad de decimales dada.
	 * @param numero El número al que se le calculará la raíz cuadrada.
	 * @param decimales La cantidad de decimales con los que se realizará el cálculo.
	 * @return Un BigDecimal con el resultado de la raíz cuadrada.
	 */
	static public BigDecimal calculeRaiz(BigDecimal numero, int decimales){
		BigDecimal diferencia, xn1, xn;
		BigDecimal dos=new BigDecimal(2);
		int precision = decimales + 6;
		int redondeo = BigDecimal.ROUND_HALF_EVEN;
		final BigDecimal cota=BigDecimal.ONE.divide(BigDecimal.TEN.pow(decimales), precision, redondeo);
		
		diferencia = xn1 = xn = BigDecimal.ONE;
		
		if (numero.compareTo(BigDecimal.ZERO) == 0) { // este es el caso en que el valor es cero
			xn1 = BigDecimal.ZERO;
		} else if (numero.compareTo(BigDecimal.ZERO) == 1) { // este es el caso en que el valor es mayor cero
			do{
				xn1 = xn.multiply(xn).add(numero).divide(xn.multiply(dos), precision, redondeo);
				diferencia = xn1.subtract(xn).abs();
				xn = xn1;
			} while (diferencia.compareTo(cota) > 0);
		}
			
		return xn1.setScale(decimales, redondeo);
	}
	
	/**
	 *Este es el procedimiento que se encarga de aproximar
	 *el valor de e^x con un número determinado de decimales
	 *de presición.
	 *@param x El valor del exponente e^x.
	 *@param decimales El número de decimales del cálculo.
	 *@return Un BigDecimal con el valor de e^x con el número de decimales dado.
	 */
	static public BigDecimal calculeE(BigDecimal x, int decimales)
	{		
		int precision = decimales + 4 + x.abs().multiply(BigDecimal.valueOf(0.44) ).intValue();	// es la precisión con la que se realizarán los cálculos
        int redondeo = BigDecimal.ROUND_HALF_EVEN;	// modo de redondeo en los cálculos
        
        // constantes que se utilizan en los cálculos
        BigDecimal aproximacion, sumando, numerador, denominador;
        int signo_x = 1, divisiones = 0;
        
        if (x.compareTo(BigDecimal.ZERO) == -1){
        	signo_x = -1;
        	x = x.negate();
        }
        
        /* 
         * se utiliza el hecho que e^x = [e^(x/2)]^2
         * se realiza x = x/2 hasta que x/2 < 1, guardando el número de divisiones realizadas en la variable
         * este cambio para el valor de x en el cálculo está directamente relacionado con el hecho de que las aproximaciones
         * utilizando esta serie de Maclaurin serán mucho mejores siempre que se consideren valores muy cercanos a cero
         */
        while (x.compareTo(BigDecimal.ONE) == 1){
        	x = x.divide(BigDecimal.valueOf(2), precision, redondeo);
        	divisiones++;
        }
        
        aproximacion = x.add(BigDecimal.ONE);
        numerador = x;
        denominador = BigDecimal.ONE;
        
        for (int i = 2; true; i++){
        	numerador = numerador.multiply(x);		// se generan x^2, x^3, x^4, ...
        	denominador = denominador.multiply(BigDecimal.valueOf(i) );		// se generan 2!, 3!, 4!, ... 
        	sumando = numerador.divide(denominador, precision, redondeo);
        	
        	if (sumando.compareTo(BigDecimal.ZERO) == 0){
    			
    			while (divisiones > 0){
    				aproximacion = aproximacion.multiply(aproximacion);
    				divisiones--;
    			}
        		
        		if (signo_x == -1)
        			return BigDecimal.ONE.divide(aproximacion, precision, redondeo).setScale(decimales, redondeo);
        		return aproximacion.setScale(decimales, redondeo);
        		
        	}
        	
        	aproximacion = aproximacion.add(sumando);
        }
	}
	
	/**
	 * Este procedimiento es el encargado de calcular el valor de pi
	 * con los decimales indicados.
	 * @param decimales La cantidad de decimales de presición deseado para calcular Pi.
	 * @return Un BigDecimal con la aproximación de Pi.
	 */
	static public BigDecimal calculePi(int decimales)
	{		
		int precision = decimales + 5;				// es la precisión con la que se realizarán los cálculos
        int redondeo = BigDecimal.ROUND_HALF_EVEN;	// modo de redondeo en los cálculos
        
        // constantes que se utilizan en la fórmula de Machin
        BigDecimal arctan1_5 = arctan(5, precision);
        BigDecimal arctan1_239 = arctan(239, precision);
        BigDecimal cuatro = new BigDecimal(4);
        
        // ejecución de la fórmula de Machin
        BigDecimal pi = arctan1_5.multiply(cuatro).subtract(arctan1_239).multiply(cuatro);
        return pi.setScale(decimales, redondeo);
    }
	
	/**
	 * Este procedimiento es el encargado de calcular el valor de arcotangente de un número
	 * con los decimales indicados cuando el número está entre -1 y 1.
	 * @param numero El valor al que se le desea calcular el arcotangente.
	 * @param presicion La cantidad de decimales de presición deseado para calcular el arcotangente.
	 * @return Un BigDecimal con la aproximación del arcotangente.
	 */
    static public BigDecimal calculeArctan(BigDecimal numero, int presicion){	
        BigDecimal resultado=numero, cociente;
        BigDecimal numeroCuadrado = numero.multiply(numero);
        BigDecimal denominador=BigDecimal.ONE, potencia=numero;
        final BigDecimal DOS=new BigDecimal(2);
        final BigDecimal cota=BigDecimal.ONE.divide((BigDecimal.TEN).pow(presicion));
        
        int i = 1;
        do{
            denominador = denominador.add(DOS);
            potencia=potencia.multiply(numeroCuadrado);
            cociente = potencia.divide(denominador, presicion, BigDecimal.ROUND_HALF_EVEN);
            
            if ((i % 2) != 0){
                resultado = resultado.subtract(cociente);
            }else{
                resultado = resultado.add(cociente);
            }
            i++;
        }while (cociente.compareTo(cota) > 0);
        
        return resultado;
    }
    
	/**
	 * Este procedimiento es el encargado de calcular el valor de arcotangente de un número
	 * con los decimales indicados cuando el número es mayor a 1 o menor a -1.
	 * @param numero El valor al que se le desea calcular el arcotangente.
	 * @param presicion La cantidad de decimales de presición deseado para calcular el arcotangente.
	 * @return Un BigDecimal con la aproximación del arcotangente.
	 */
    static public BigDecimal calculeArctanMayorMenor(BigDecimal numero, int presicion){	
        BigDecimal valorInicial=new BigDecimal(1.57);
        BigDecimal seno, coseno;
        BigDecimal valorSiguiente;
        BigDecimal diferencia;
        //BigDecimal valorMayor=new BigDecimal(1.57079632679);
        boolean negativo=false;
        final BigDecimal cota=BigDecimal.ONE.divide((BigDecimal.TEN).pow(presicion));
        
        if(numero.compareTo(BigDecimal.ZERO)<0){
        	numero=numero.negate();
        	negativo=true;
        }
        
        do{
        	seno=calculeSeno(valorInicial, presicion);
        	coseno=calculeCoseno(valorInicial, presicion);
        	valorSiguiente=(new BigDecimal(-1)).multiply(seno.multiply(coseno)).add(numero.multiply(coseno.multiply(coseno))).add(valorInicial);
            /*if(valorSiguiente.compareTo(valorMayor)>0){
            	valorSiguiente=new BigDecimal("1.5707963267948966192313216916397514420985846996876");
            }*/
            diferencia=valorInicial.subtract(valorSiguiente).abs();
            valorInicial=valorSiguiente;
            
        }while (diferencia.compareTo(cota) > 0);
        
        if(negativo){
        	valorSiguiente=valorSiguiente.negate();
        }
        
        return valorSiguiente;
    }
    
    /*
     *este procedimiento aproxima el valor de arctan(x) utilizando
     *arctan(x) = x - x^3/3 + x^5/5 - x^7/7 + x^9/9 - x^11/11 + ...
     *para valores que satisfagan |x| <= 1
     */
    static private BigDecimal arctan(int inverso_x, int precision)
    {	
		int redondeo = BigDecimal.ROUND_HALF_EVEN;	// modo de redondeo en los cálculos
        
        // variables que se utilizarán en el procedimiento
        // resulado:	guarda el valor de nuestra aproximación de pi
        // numero:		es el valor al que se le aproximará su arctan; en este caso, será 1/5 y 1/239
        // sumando:		lleva el valor de los términos de la serie que aproxima arctan(x)
        BigDecimal resultado, numero, sumando;
        BigDecimal cuadrado_inverso_x = BigDecimal.valueOf(inverso_x * inverso_x);

        numero = BigDecimal.ONE.divide(BigDecimal.valueOf(inverso_x), precision, redondeo);
        resultado = numero;
        int i = 1;
        
        do{
            numero = numero.divide(cuadrado_inverso_x, precision, redondeo);
            int denominador = 2 * i + 1;
            sumando = numero.divide(BigDecimal.valueOf(denominador), precision, redondeo);
            
            if ((i % 2) != 0)
                resultado = resultado.subtract(sumando);
            else
                resultado = resultado.add(sumando);
            i++;
        }while (sumando.compareTo(BigDecimal.ZERO) != 0);
        
        return resultado;
    }
}
