package itcr.mathexplorer.herramientas;

import java.math.*;

/**
 * @(#)Punto3D.java
 * 
 *Clase para guardar un punto en tres dimensiones.
 *
 *@author Alexander Borbón Alpízar
 */

public class Punto3D{
	
	//VARIABLES PUBLICAS
	
	/**
	 * El Punto (0,0,0)
	 */
	final public static Punto3D ZERO=new Punto3D(BigDecimal.ZERO, BigDecimal.ZERO, BigDecimal.ZERO);
	
	/**
	 * El Punto (1,0,0)
	 */
	final public static Punto3D X=new Punto3D(BigDecimal.ONE, BigDecimal.ZERO, BigDecimal.ZERO);
	
	/**
	 * El Punto (0,1,0)
	 */
	final public static Punto3D Y=new Punto3D(BigDecimal.ZERO, BigDecimal.ONE, BigDecimal.ZERO);
	
	/**
	 * El Punto (0,0,1)
	 */
	final public static Punto3D Z=new Punto3D(BigDecimal.ZERO, BigDecimal.ZERO, BigDecimal.ONE);
	
	//VARIABLES PRIVADAS
	
	private BigDecimal x, y, z;
	
	/**
	 *Constructor que crea el punto a partir de números decimales grandes.
	 *@param a La coordenada x del punto.
	 *@param b La coordenada y del punto.
	 *@param c La coordenada z del punto.
	 */
	public Punto3D(BigDecimal a, BigDecimal b, BigDecimal c){
		x=a;
		y=b;
		z=c;
	}
	
	/**
	 *Constructor que crea el punto a partir de números enteros.
	 *@param a La coordenada x del punto.
	 *@param b La coordenada y del punto.
	 *@param c La coordenada z del punto.
	 */
	public Punto3D(float a, float b, float c){
		x=new BigDecimal(a);
		y=new BigDecimal(b);
		z=new BigDecimal(c);
	}
	
	/**
	 *Constructor que crea el punto a partir de números enteros.
	 *@param a La coordenada x del punto.
	 *@param b La coordenada y del punto.
	 *@param c La coordenada z del punto.
	 */
	public Punto3D(int a, int b, int c){
		x=new BigDecimal(a);
		y=new BigDecimal(b);
		z=new BigDecimal(c);
	}
	
	/**
	 * Función que retorna la coordenada x del punto.
	 * @return Retorna la coordenada x del punto.
	 */
	public BigDecimal getX(){
		return x;
	}
	
	/**
	 * Función que retorna la coordenada y del punto.
	 * @return Retorna la coordenada y del punto.
	 */
	public BigDecimal getY(){
		return y;
	}
	
	/**
	 * Función que retorna la coordenada z del punto.
	 * @return Retorna la coordenada z del punto.
	 */
	public BigDecimal getZ(){
		return z;
	}
	
	/**
	 * Función que le suma a este punto otro punto en 3D.
	 * @param A El punto a sumar.
	 * @return Un nuevo Punto3D que contiene la suma de este punto y A.
	 */
	public Punto3D suma(Punto3D A){
		return new Punto3D(this.getX().add(A.getX()), this.getY().add(A.getY()), this.getZ().add(A.getZ()));
	}
	
	/**
	 * Función que multiplica este punto por una constante.
	 * @param c La constante a multiplicar.
	 * @return Un nuevo Punto3D que contiene la multiplicación de c por este punto.
	 */
	public Punto3D multiplica(BigDecimal c){
		return new Punto3D(c.multiply(this.getX()), c.multiply(this.getY()), c.multiply(this.getZ()));
	}
	
	/**
	 * Función que divide este punto por una constante.
	 * @param c La constante a dividir.
	 * @return Un nuevo Punto3D que contiene la divición de este punto por c.
	 */
	public Punto3D divide(BigDecimal c){
		return new Punto3D(this.getX().divide(c,2), this.getY().divide(c,2), this.getZ().divide(c,2));
	}
	
	/**
	 * Función que calcula el producto punto de este punto por A.
	 * @param A El punto a calcular el producto punto.
	 * @return Un valor BigDecimal que contiene el producto punto de este punto por A.
	 */
	public BigDecimal productoPunto(Punto3D A){
		return this.getX().multiply(A.getX()).add(this.getY().multiply(A.getY())).add(this.getZ().multiply(A.getZ()));
	}
	
	/**
	 * Función que calcula el producto cruz de este punto por A.
	 * @param A El punto a calcular el producto cruz.
	 * @return Un nuevo Punto3D que contiene el producto cruz de este punto por A.
	 */
	public Punto3D productoCruz(Punto3D A){
		return new Punto3D(this.getY().multiply(A.getZ()).subtract(this.getZ().multiply(A.getY())),this.getZ().multiply(A.getX()).subtract(this.getX().multiply(A.getZ())), this.getX().multiply(A.getY()).subtract(this.getY().multiply(A.getX())));
	}
	
	/**
	 * Función que calcula la norma del vector representado por este punto.
	 * @param decimales El número de decimales con los que se calculará la norma.
	 * @return La norma del vector que representa este punto.
	 */
	public BigDecimal norma(int decimales){
		return MetodosNumericos.calculeRaiz(this.getX().multiply(this.getX()).add(this.getY().multiply(this.getY())).add(this.getZ().multiply(this.getZ())), 2);
	}
	
	/**
	 * Redondea este punto con la cantidad de decimales dada.
	 * @param decimales La cantidad de decimales a redondear.
	 * @return El nuevo punto redondeado
	 */
	public Punto3D redondear(int decimales){
		return new Punto3D(this.x.round(new MathContext(decimales)), this.y.round(new MathContext(decimales)), this.z.round(new MathContext(decimales)));
	}
}