package itcr.mathexplorer.traductores;

import java.util.*;
import itcr.mathexplorer.exceptions.*;
import itcr.mathexplorer.tablas.*;
import itcr.mathexplorer.tipos.*;
import itcr.mathexplorer.expresion.*;
import java.math.BigInteger;
import java.math.BigDecimal;

/**
 *@(#)TraductorInfijoPrefijo.java
 *
 *Clase que traduce una expresión matemática a prefijo.
 *Para traducir se debe crear un objeto (Obj) de tipo TraductorPrefijo.
 *Para parsear expr se escribe Obj.traducir(Expr).
 *La función devuelve un Nodo con Expr en notación prefija en un árbol.
 *
 *Se permite cualquier variable
 *La expresión puede contener las constantes #pi y #e.
 *Los operadores, funciones y constantes válidos por defecto para la expresión son:
 *    OPERACIÓN           OPERADOR
 *  suma                    +
 *  resta                   -
 *  multiplicación          *
 *  división                /
 *  potencias               ^
 *  módulo                  %
 *  paréntesis              ( )
 *  listas                  { }
 *  logaritmo (base e)      ln( )
 *  logaritmo (base 10)     log( )
 *  valor absoluto          abs( )
 *  número aleatorio        rnd( )
 *  seno                    sen( )
 *  coseno                  cos( )
 *  tangente                tan( )
 *  secante                 sec( )
 *  cosecante               csc( )
 *  cotangente              cot( )
 *  arcoseno                asen( )
 *  arcocoseno              acos( )
 *  arcotangente            atan( )
 *  arcosecante             asec( )
 *  arcocosecante           acsc( )
 *  arcocotangente          acot( )
 *  seno hiperbólico        senh( )
 *  coseno hiperbólico      cosh( )
 *  tangente hiperbólica    tanh( )
 *  secante hiperbólica     sech( )
 *  cosecante hiperbólica   csch( )
 *  cotangente hiperbólica  coth( )
 *  raices cuadradas        sqrt( )
 *  redondeo                round( )
 *
 *Algunos ejemplos de expresiones válidas son:
 *x+cos(3)*tan(x^(2*#pi*x-1))/acos(1/2)
 *cosh(x)+abs(1-x^2)%3
 *{{1,2},x,y}
 *
 *@author Alexander Borbón Alpízar
 */
public class TraductorInfijoPrefijo{

	//VARIABLES PRIVADAS
	
	private int maxNumCar;
	private TablaOperadores operadores;
	private TablaFunciones funciones;
	private TablaConstantes constantes;
	private TablaFuncionesUsuario funcionesUsuario;
	//private int numDecimales=2;
	
	
	//CONSTRUCTORES
	
	/**
	 *El constructor de la expresión, debe recibir las tablas con las funciones con las que se trabajará.
	 *@param tablaoperadores La tabla que contiene todos los operadores disponibles para la expresión.
	 *@param tablafunciones La tabla que contiene todas las funciones disponibles para la expresión.
	 *@param tablaconstantes La tabla que contiene todas las constantes disponibles para la expresión.
	 *@param tablafuncionesusuario La tabla que contiene todas las funciones definidas por el usuario disponibles para la expresión.
	 */
	public TraductorInfijoPrefijo(TablaOperadores tablaoperadores, TablaFunciones tablafunciones, TablaConstantes tablaconstantes, TablaFuncionesUsuario tablafuncionesusuario){
		operadores=tablaoperadores;
		funciones=tablafunciones;
		constantes=tablaconstantes;
		funcionesUsuario=tablafuncionesusuario;
		maxNumCar=obtenerMaxNumCar();
	}


	//FUNCIONES PUBLICAS
	
	/**
	 *Devuelve el número de caracteres que tiene la función, constante u operador mayor en la expresión.
	 *@return El máximo número de caracteres en la expresión.
	 */
	public int getMaxNumCar(){
		return maxNumCar;
	}
	
	/*
	 *Define el número de decimales con los que se está trabajando, por defecto este número es dos.
	 *@param nuevoNumDecimales El nuevo número de decimales a utilizar, este debe ser un número positivo.
	 *@exception IllegalArgumentException Se lanza si el número de decimales es negativo o cero.
	 */
	/*public void setNumDecimales(int nuevoNumDecimales) throws IllegalArgumentException{
		if(nuevoNumDecimales>0){
			numDecimales=nuevoNumDecimales;
		}else{
			throw new IllegalArgumentException("El número de decimales no puede ser negativo");
		}
	}*/
	
	/**
	 *La función que traduce la expresión a notación prefijo.
	 *@param expresion El string con la expresión a traducir.
	 *@return Un Nodo con la expresión traducida en notación prefijo.
	 *@exception SintaxException Error de escritura en la expresión.
	 */
	public Nodo traducir(String expresion) throws SintaxException{
		Stack<Nodo> PilaNumeros=new Stack<Nodo>(); //Pila donde se guardarán los números al parsear
		Stack<Nodo> PilaOperadores= new Stack<Nodo>(); //Pila donde se guardarán los operadores al parsear
		String expr=quitaEspacios(expresion.toLowerCase());  //La expresión sin espacios ni mayúsculas.
		String pedazo, numero;
		Nodo comodin, comodin2, expresiontraducida;
		BigInteger entero;
		BigDecimal decimal;
		boolean esDecimal, salir, numeronegativo=false;
		int pos=0, i=0;
		/*
		 *Esta variable 'anterior' se utiliza para saber cuál había sido la última
		 *cosa parseada, se utiliza:
		 *0 : no ha parseado nada
		 *1 : un número o variable: número, pi, e, x, rnd
		 *2 : un operador binario: + - * / ^ %
		 *3 : un paréntesis: { ( sen( cos( etc.
		 *4 : cierre de paréntesis: ) }
		 *5 : una coma
		 *Si no se ha parseado nada puede ser cualquier cosa menos: + * / ^ %
		 *Si el anterior fue un número puede seguir cualquier cosa
		 *Si lo anterior fue un operador puede seguir cualquier cosa menos otro operador (con excepción de -)
		 *Si lo anterior fue un paréntesis puede seguir cualquier cosa menos: + * / ^ %
		 *Si lo anterior fue un cierre de paréntesis debe seguir un operador, un número (en cuyo caso hay un por oculto) u otro paréntesis (también hay un por oculto)
		 */
		byte anterior=0;
		
		try{
			
			//Procedimiento para trabajar la definición de funciones y variables (:=)
			
			int asignacion=expr.indexOf(":=");
			if(asignacion!=-1){
				if(expr.substring(asignacion+2).indexOf(":=")!=-1){
					throw new SintaxException("No se permite la doble asignación.");
				}else{
					char c=expr.charAt(pos);
					if(isNum(c)){
						throw new SintaxException("No se permite un nombre de asignación que inicie con número.");
					}else{
						
						//Este procedimiento saca el nombre de la función
						String fun=new String();
						while(c!='(' && c!=':'){
							if(!(isNum(c) || isLetra(c))){
								throw new SintaxException("Sólo se permiten números y letras en los nombres de asignación.");
							}
							fun=fun+c;
							
							pos+=1;
							c=expr.charAt(pos);
						}
						
						comodin=new Nodo(fun);
						
						if(c=='('){
							pos+=1;
							c=expr.charAt(pos);
							
							while(c!=':'){
								fun="";
								while(c!=',' && c!=')'){
									if(!(isNum(c) || isLetra(c))){
										return new Nodo("Indefinido");
									}
									fun=fun+c;
									
									pos+=1;
									c=expr.charAt(pos);
								}
								if(!fun.equals("")){
									comodin2=new Nodo("Variable");
									comodin2.agregaOperando(new Nodo(fun));
									comodin.agregaOperando(comodin2);
								}
								pos+=1;
								c=expr.charAt(pos);
							}
						}
						
						comodin2=new Nodo(":=");
						
						comodin2.agregaOperando(comodin);
						
						Nodo comodin3=traducir(expr.substring(asignacion+2));
						
						comodin2.agregaOperando(comodin3);
						
						funcionesUsuario.agregaFuncion(comodin.getTipo(),new FuncionUsuario(comodin.clonarNodo(), comodin3.clonarNodo()));
						
						return comodin2;
					}
				}
			}
			
			//Maneja todo el resto de casos
			
			while(pos<expr.length()){
				if(isNum(expr.charAt(pos))){ //Si lo que hay que sacar es un número
					if(anterior==1 || anterior==4){ //Si lo que hay antes es un número o variable o cierre de paréntesis es porque hay una multiplicación oculta.
						//Se saca todo lo que está con prioridad menor que la multiplicación
						while(!PilaOperadores.empty() && operadores.apareceLlave(((Nodo)PilaOperadores.peek()).getTipo()) && operadores.devuelveOperador(((Nodo)PilaOperadores.peek()).getTipo()).prioridad()>=operadores.devuelveOperador("·").prioridad()){
							sacaOperador(PilaNumeros, PilaOperadores);
						}
						PilaOperadores.push(new Nodo("·"));
					}
					
					if(numeronegativo){
						pedazo="-";
						numeronegativo=false;
					}else{
						pedazo="";
					}
					esDecimal=false;
					while(pos<expr.length() && (isNum(expr.charAt(pos)) || expr.charAt(pos) == '.')){
						if (expr.charAt(pos) == '.'){
							esDecimal=true;
						}
						pedazo=pedazo+expr.charAt(pos);
						pos++;
					}
					
					try{
						if (esDecimal){
							decimal=new BigDecimal(pedazo);
							numero=decimal.toString();
							if(numero.length()<=18){
								comodin=new Nodo("Decimal",1);
							}else{
								comodin=new Nodo("DecimalLargo",1);
							}
							comodin.agregaOperando(new Nodo(numero));
							PilaNumeros.push(comodin);
						}else{
							entero=new BigInteger(pedazo);
							numero=entero.toString();
							if(numero.length()<=18){
								comodin=new Nodo("Entero",1);
							}else{
								comodin=new Nodo("EnteroLargo",1);
							}
							comodin.agregaOperando(new Nodo(numero));
							PilaNumeros.push(comodin);
						}
					}catch(NumberFormatException e){
						throw new SintaxException("Número mal digitado");
					}
					anterior=1;
				}else{ //Si lo que hay que sacar NO es un número
				
					salir=false;
					i=maxNumCar+1;
					
					while(i>0 && !salir){
						
						if(pos+i<=expr.length()){
							
							pedazo=expr.substring(pos,pos+i);
							
							if(operadores.apareceLlave(pedazo)){ //Si es un operador
								if(anterior==0 || anterior==2 || anterior==3 || anterior==5){ //No ha parseado nada, es un operador o un paréntesis o una coma
									if(pedazo.equals("-")){ //Es un menos unario
										
										if(pos+i+1<=expr.length() && isNum(expr.charAt(pos+i))){
											numeronegativo=true;
										}else{
											comodin=new Nodo("Entero",1);
											comodin.agregaOperando(new Nodo("-1"));
											PilaNumeros.push(comodin);
											/*while(!PilaOperadores.empty() && operadores.apareceLlave(((Nodo)PilaOperadores.peek()).getTipo()) && operadores.devuelveOperador(((Nodo)PilaOperadores.peek()).getTipo()).prioridad()>=operadores.devuelveOperador("·").prioridad()){
												sacaOperador(PilaNumeros, PilaOperadores);
											}*/
											PilaOperadores.push(new Nodo("·"));
										}
									}else if(!operadores.devuelveOperador(pedazo).esBinario()){ //Es un operador unario distinto de -
										comodin=new Nodo(pedazo,1);
										PilaOperadores.push(comodin);
									}else{
										throw new SintaxException("Error en la expresión");
									}
								}else{
									if(pedazo.equals("-")){
										//Primero saca los anteriores según su procedencia
										while(!PilaOperadores.empty() && operadores.apareceLlave(((Nodo)PilaOperadores.peek()).getTipo()) && operadores.devuelveOperador(((Nodo)PilaOperadores.peek()).getTipo()).prioridad()>=operadores.devuelveOperador(pedazo).prioridad()){
											sacaOperador(PilaNumeros, PilaOperadores);
										}
										
										//Luego mete un "+" como operador y un "-1" y un "·", es decir convierte el "-" en "+ -1*"
										PilaOperadores.push(new Nodo("+"));
										
										if(pos+i+1<=expr.length() && isNum(expr.charAt(pos+i))){
											numeronegativo=true;
										}else{
											comodin=new Nodo("Entero",1);
											comodin.agregaOperando(new Nodo("-1"));
											PilaNumeros.push(comodin);
											
											//Como lo anterior fue un "-" el "·" no lo va a sacar
											/*while(!PilaOperadores.empty() && operadores.apareceLlave(((Nodo)PilaOperadores.peek()).getTipo()) && operadores.devuelveOperador(((Nodo)PilaOperadores.peek()).getTipo()).prioridad()>=operadores.devuelveOperador("·").prioridad()){
												sacaOperador(PilaNumeros, PilaOperadores);
											}*/
											PilaOperadores.push(new Nodo("·"));
										}
									}else if (pedazo.equals("^")){
										if(anterior==1 && ((Nodo)PilaNumeros.peek()).getOperando(1).getTipo().indexOf("-")!=-1){ //Si el anterior es un número y es negativo entonces el menos no queda elevado a la potencia
											comodin=(Nodo)PilaNumeros.pop();
											
											comodin2=new Nodo("Entero",1);
											comodin2.agregaOperando(new Nodo("-1"));
											PilaNumeros.push(comodin2);
											
											//Hay que sacar todo lo necesario con respecto al *
											while(!PilaOperadores.empty() && operadores.apareceLlave(((Nodo)PilaOperadores.peek()).getTipo()) && operadores.devuelveOperador(((Nodo)PilaOperadores.peek()).getTipo()).prioridad()>=operadores.devuelveOperador("·").prioridad()){
												sacaOperador(PilaNumeros, PilaOperadores);
											}
											PilaOperadores.push(new Nodo("·"));
											
											numero=comodin.getOperando(1).getTipo().substring(1);
											comodin.eliminaOperando(1);
											comodin.agregaOperando(new Nodo(numero));
											PilaNumeros.push(comodin);
										}
										
										if(anterior!=4){ //Si no hay un paréntesis de cierre antes
											while(!PilaOperadores.empty() && operadores.apareceLlave(((Nodo)PilaOperadores.peek()).getTipo()) && operadores.devuelveOperador(((Nodo)PilaOperadores.peek()).getTipo()).prioridad()>operadores.devuelveOperador(pedazo).prioridad()){
												sacaOperador(PilaNumeros, PilaOperadores);
											}
										}
										PilaOperadores.push(new Nodo(pedazo));
									}else if(pedazo.equals("!")){//Si es un factorial (este se maneja de manera distinta al resto de operadores)
										comodin=new Nodo("!");
										comodin.agregaOperando((Nodo)PilaNumeros.pop());
										PilaNumeros.push(comodin);
									}else if(pedazo.equals("%")){//Si es un porcentaje (este se maneja de manera distinta al resto de operadores)
										comodin=new Nodo("%");
										comodin.agregaOperando((Nodo)PilaNumeros.pop());
										PilaNumeros.push(comodin);
									}else{
										if(pedazo.equals("*")){
											pedazo="·";
										}
										
										while(!PilaOperadores.empty() && operadores.apareceLlave(((Nodo)PilaOperadores.peek()).getTipo()) && operadores.devuelveOperador(((Nodo)PilaOperadores.peek()).getTipo()).prioridad()>=operadores.devuelveOperador(pedazo).prioridad()){
											sacaOperador(PilaNumeros, PilaOperadores);
										}
										PilaOperadores.push(new Nodo(pedazo));
									}
								}
						
								salir=true;
								anterior=2;
								if(pedazo.equals("!") || pedazo.equals("%")){//Si es un factorial o un porcentaje se trata como si fuera un número
									anterior=1;
								}
							}else if(constantes.apareceLlave(pedazo)){ //Si es una constante
								if(anterior==1 || anterior==4){ //Si lo que hay antes es un número o variable o cierre de paréntesis es porque hay una multiplicación oculta.
									//Se saca todo lo que está con prioridad menor que la multiplicación
									while(!PilaOperadores.empty() && operadores.apareceLlave(((Nodo)PilaOperadores.peek()).getTipo()) && operadores.devuelveOperador(((Nodo)PilaOperadores.peek()).getTipo()).prioridad()>=operadores.devuelveOperador("·").prioridad()){
										sacaOperador(PilaNumeros, PilaOperadores);
									}
									PilaOperadores.push(new Nodo("·"));
								}
								comodin=new Nodo("Constante",1);
								comodin.agregaOperando(new Nodo(expr.substring(pos,pos+i)));
								PilaNumeros.push(comodin);
								
								salir=true;
								anterior=1;
							}else if (pedazo.equals("(")){ //Si es un paréntesis de apertura
								if(anterior==1 || anterior==4){ //Si lo que hay antes es un número o variable o cierre de paréntesis es porque hay una multiplicación oculta.
									//Se saca todo lo que está con prioridad menor que la multiplicación
									while(!PilaOperadores.empty() && operadores.apareceLlave(((Nodo)PilaOperadores.peek()).getTipo()) && operadores.devuelveOperador(((Nodo)PilaOperadores.peek()).getTipo()).prioridad()>=operadores.devuelveOperador("·").prioridad()){
										sacaOperador(PilaNumeros, PilaOperadores);
									}
									PilaOperadores.push(new Nodo("·"));
								}
								PilaOperadores.push(new Nodo("("));
								
								salir=true;
								anterior=3;
							}else if(expr.substring(pos+i-1,pos+i).equals("(") && (funciones.apareceLlave(expr.substring(pos,pos+i-1)) || funcionesUsuario.apareceLlave(expr.substring(pos,pos+i-1)))){ //Si es una función (definida por el usuario o de las normales)
								if(anterior==1 || anterior==4){ //Si lo que hay antes es un número o variable o cierre de paréntesis es porque hay una multiplicación oculta.
									//Se saca todo lo que está con prioridad menor que la multiplicación
									while(!PilaOperadores.empty() && operadores.apareceLlave(((Nodo)PilaOperadores.peek()).getTipo()) && operadores.devuelveOperador(((Nodo)PilaOperadores.peek()).getTipo()).prioridad()>=operadores.devuelveOperador("·").prioridad()){
										sacaOperador(PilaNumeros, PilaOperadores);
									}
									PilaOperadores.push(new Nodo("·"));
								}
								PilaOperadores.push(new Nodo(expr.substring(pos,pos+i-1)));
								
								salir=true;
								anterior=3;
							}else if(funcionesUsuario.apareceLlave(pedazo) && ((FuncionUsuario)(funcionesUsuario.devuelveFuncion(pedazo))).obtenFuncion().getNumeroOperandos()==0){ //Si es una variable definida por el usuario
								if(anterior==1 || anterior==4){ //Si lo que hay antes es un número o variable o cierre de paréntesis es porque hay una multiplicación oculta.
									//Se saca todo lo que está con prioridad menor que la multiplicación
									while(!PilaOperadores.empty() && operadores.apareceLlave(((Nodo)PilaOperadores.peek()).getTipo()) && operadores.devuelveOperador(((Nodo)PilaOperadores.peek()).getTipo()).prioridad()>=operadores.devuelveOperador("·").prioridad()){
										sacaOperador(PilaNumeros, PilaOperadores);
									}
									PilaOperadores.push(new Nodo("·"));
								}
								PilaNumeros.push(new Nodo(pedazo));
								
								salir=true;
								anterior=1;
							}else if (pedazo.equals(")")){ //Si es un parentesis de cierre
								if(anterior!=1 && anterior!=3 && anterior!=4){ //Antes de un paréntesis de cierre sólo puede haber un número, un paréntesis de apertura u otro paréntesis de cierre.
									throw new SintaxException("Error en algún paréntesis de cierre");
								}
								
								while(!PilaOperadores.empty() && !funcionesUsuario.apareceLlave(((Nodo)PilaOperadores.peek()).getTipo()) && !funciones.apareceLlave(((Nodo)PilaOperadores.peek()).getTipo()) && !(((Nodo)PilaOperadores.peek()).getTipo()).equals("(")){
									sacaOperador(PilaNumeros, PilaOperadores);
								}
								
								if((((Nodo)PilaOperadores.peek()).getTipo()).equals("(")){
									PilaOperadores.pop();
								}else{
									comodin=(Nodo)PilaOperadores.pop();
									if(anterior==1 || anterior==4){
										comodin.agregaOperando((Nodo)PilaNumeros.pop());
									}
									
									if(funciones.apareceLlave(comodin.getTipo()) && funciones.devuelveFuncion(comodin.getTipo()).permiteNumeroParametros(comodin.getNumeroOperandos())){
											PilaNumeros.push(comodin);
									}else if(funcionesUsuario.apareceLlave(comodin.getTipo()) && funcionesUsuario.devuelveFuncion(comodin.getTipo()).permiteNumeroParametros(comodin.getNumeroOperandos())){
											PilaNumeros.push(comodin);
									}else{
										throw new SintaxException("La función no admite el número de argumentos dado.");
									}
								}
								
								salir=true;
								anterior=4;
							}else if (pedazo.equals("{")){ //Si es una llave de apertura
								if(anterior==1 || anterior==4){ //Si lo que hay antes es un número o variable o cierre de paréntesis es porque hay una multiplicación oculta.
									//Se saca todo lo que está con prioridad menor que la multiplicación
									while(!PilaOperadores.empty() && operadores.apareceLlave(((Nodo)PilaOperadores.peek()).getTipo()) && operadores.devuelveOperador(((Nodo)PilaOperadores.peek()).getTipo()).prioridad()>=operadores.devuelveOperador("·").prioridad()){
										sacaOperador(PilaNumeros, PilaOperadores);
									}
									PilaOperadores.push(new Nodo("·"));
								}
								PilaOperadores.push(new Nodo("Lista"));
								
								salir=true;
								anterior=3;
							}else if (pedazo.equals("}")){ //Si es una llave de cierre
								if(anterior!=1 && anterior!=3 && anterior!=4){ //Antes de un paréntesis de cierre sólo puede haber un número, un paréntesis de apertura u otro paréntesis de cierre.
									throw new SintaxException("Error en alguna llave de cierre");
								}
								
								while(!PilaOperadores.empty() && !(((Nodo)PilaOperadores.peek()).getTipo()).equals("Lista")){
									sacaOperador(PilaNumeros, PilaOperadores);
								}
								
								comodin=(Nodo)PilaOperadores.pop();
								if(anterior==1 || anterior==4){
									comodin.agregaOperando((Nodo)PilaNumeros.pop());
								}
								PilaNumeros.push(comodin);
								
								salir=true;
								anterior=4;
							}else if (pedazo.equals("[")){ //Si es el inicio de una matriz
								if(anterior==1 || anterior==4){ //Si lo que hay antes es un número o variable o cierre de paréntesis es porque hay una multiplicación oculta.
									//Se saca todo lo que está con prioridad menor que la multiplicación
									while(!PilaOperadores.empty() && operadores.apareceLlave(((Nodo)PilaOperadores.peek()).getTipo()) && operadores.devuelveOperador(((Nodo)PilaOperadores.peek()).getTipo()).prioridad()>=operadores.devuelveOperador("·").prioridad()){
										sacaOperador(PilaNumeros, PilaOperadores);
									}
									PilaOperadores.push(new Nodo("·"));
								}
								PilaOperadores.push(new Nodo("Matriz"));
								PilaOperadores.push(new Nodo("Fila"));
								
								salir=true;
								anterior=3;
							}else if (pedazo.equals("]")){ //Si es el fin de una matriz
								if(anterior!=1 && anterior!=3 && anterior!=4){ //Antes de un punto coma sólo puede haber un número, un paréntesis de cierre.
									throw new SintaxException("Error al final de una matriz");
								}
								
								while(!PilaOperadores.empty() && !(((Nodo)PilaOperadores.peek()).getTipo()).equals("Fila") && !((Nodo)PilaOperadores.peek()).getTipo().equals("Matriz")){ 
									sacaOperador(PilaNumeros, PilaOperadores);
								}
								
								comodin=(Nodo)PilaOperadores.pop();
								if(anterior==1 || anterior==4){
									comodin.agregaOperando((Nodo)PilaNumeros.pop());
								}
								comodin2=(Nodo)PilaOperadores.pop();
								if(comodin.getNumeroOperandos()!=0){
									comodin2.agregaOperando(comodin);
								}
								
								for(int j=2; j<=comodin2.getNumeroOperandos(); j++){
									if(comodin2.getOperando(j-1).getNumeroOperandos()!=comodin2.getOperando(j).getNumeroOperandos()){
										throw new SintaxException("El número de elementos por fila en la matriz no es igual");
									}
								}
								
								PilaNumeros.push(comodin2);
								
								salir=true;
								anterior=4;
							}else if (pedazo.equals(",")){
								if(anterior!=1 && anterior!=4){ //Antes de una coma sólo puede haber un número, un paréntesis de cierre.
									throw new SintaxException("Error en alguna coma");
								}
								
								while(!PilaOperadores.empty() && !funciones.apareceLlave(((Nodo)PilaOperadores.peek()).getTipo()) && !funcionesUsuario.apareceLlave(((Nodo)PilaOperadores.peek()).getTipo()) && !(((Nodo)PilaOperadores.peek()).getTipo()).equals("(") && !(((Nodo)PilaOperadores.peek()).getTipo()).equals("Fila") && !((Nodo)PilaOperadores.peek()).getTipo().equals("Lista")){ 
									sacaOperador(PilaNumeros, PilaOperadores);
								}
								
								if((((Nodo)PilaOperadores.peek()).getTipo()).equals("(")){
									throw new SintaxException("Error en alguna coma");
								}else{
									comodin=(Nodo)PilaOperadores.pop();
									comodin.agregaOperando((Nodo)PilaNumeros.pop());
									PilaOperadores.push(comodin);
								}
								
								salir=true;
								anterior=5;
								
							}else if (pedazo.equals(";")){
								if(anterior!=1 && anterior!=4){ //Antes de un punto coma sólo puede haber un número, un paréntesis de cierre.
									throw new SintaxException("Error en algun punto y coma");
								}
								
								while(!PilaOperadores.empty() && !(((Nodo)PilaOperadores.peek()).getTipo()).equals("Fila")){ // && !((Nodo)PilaOperadores.peek()).getTipo().equals("Matriz")){ 
									sacaOperador(PilaNumeros, PilaOperadores);
								}
								
								comodin=(Nodo)PilaOperadores.pop();
								comodin.agregaOperando((Nodo)PilaNumeros.pop());
								comodin2=(Nodo)PilaOperadores.pop();
								comodin2.agregaOperando(comodin);
								PilaOperadores.push(comodin2);
								PilaOperadores.push(new Nodo("Fila"));
								
								salir=true;
								anterior=5;
								
							}else if(i==1){ //si i=1 y no entró en ninguno de los casos anteriores es porque debe ser una variable desconocida.
								if(anterior==1 || anterior==4){ //Si lo que hay antes es un número o variable o cierre de paréntesis es porque hay una multiplicación oculta.
									//Se saca todo lo que está con prioridad menor que la multiplicación
									while(!PilaOperadores.empty() && operadores.apareceLlave(((Nodo)PilaOperadores.peek()).getTipo()) && operadores.devuelveOperador(((Nodo)PilaOperadores.peek()).getTipo()).prioridad()>=operadores.devuelveOperador("·").prioridad()){
										sacaOperador(PilaNumeros, PilaOperadores);
									}
									PilaOperadores.push(new Nodo("·"));
								}
								comodin=new Nodo("Variable",1);
								comodin.agregaOperando(new Nodo(expr.substring(pos,pos+i)));
								PilaNumeros.push(comodin);
								
								salir=true;
								anterior=1;
							}
						}
						
						i--;
					}

					pos+=i+1;
				}
			}
		
			//Procesa al final
			while(!PilaOperadores.empty()){
				sacaOperador(PilaNumeros, PilaOperadores);
			}
		
		}catch(EmptyStackException e){
			throw new SintaxException("Expresión mal digitada");
		}
		
		
		expresiontraducida=((Nodo)PilaNumeros.pop());
		
		if(!PilaNumeros.empty()){
			throw new SintaxException("Error en la expresión");
		}
		
		return expresiontraducida;
	}//Parsear
	

	
	//FUNCIONES PRIVADAS
	
	
	/*
	 *sacaOperador es una función que se encarga de sacar un operador de la pila
	 *y manejarlo de manera apropiada, ya sea un operador unario o binario
	 */
	private void sacaOperador(Stack<Nodo> PilaNumeros, Stack PilaOperadores) throws EmptyStackException{
		Nodo operador;
		Nodo a, b;
		
		try{
			operador=(Nodo)PilaOperadores.pop();
			
			if(operadores.devuelveOperador(operador.getTipo()).esBinario()){
				b=(Nodo)PilaNumeros.pop();
				a=(Nodo)PilaNumeros.pop();
				operador.agregaOperando(a);
				operador.agregaOperando(b);
				PilaNumeros.push(operador);
			}else{
				a=(Nodo)PilaNumeros.pop();
				operador.agregaOperando(a);
				PilaNumeros.push(operador);
			}
		}catch(EmptyStackException e){
			e.printStackTrace();
			throw e;
		}
	}//sacaOperador
	
	/*
	 *Función que pregunta si un caracter es un número o no
	 */
	private boolean isNum(char s) {
		if (s >= '0' && (s <= '9'))
			return true;
		else
			return false;
	} //Fin de la funcion isNum
	
	/*
	 *Función que pregunta si un caracter es una letra de la 'a' a la 'z'
	 */
	private boolean isLetra(char s) {
		if (s >= 'a' && (s <= 'z'))
			return true;
		else
			return false;
	} //Fin de la funcion isLetra
	
	/*
	 *Quita los espacios del String
	 */	
	private String quitaEspacios(String cadena){
		String unspacedString = "";	//Variable donde lee la función

		for(int i = 0; i < cadena.length(); i++){	//Le quita los espacios a la expresión que leyó
			if(cadena.charAt(i) != ' ')
				unspacedString += cadena.charAt(i);
		}//for
		
		return unspacedString;
	}//quitaEspacios
	
	/*
	 *Esta función obtiene el máximo número de caracteres que puede tener una función, un caracter o un operador
	 */
	private int obtenerMaxNumCar(){
		int max=0;
		Object op[], fu[], co[];
		op=operadores.keySet().toArray();
		fu=funciones.keySet().toArray();
		co=constantes.keySet().toArray();
		
		for(int i=0;i<op.length;i++){
			max=java.lang.Math.max(max,((String)op[i]).length());
		}
		
		for(int i=0;i<fu.length;i++){
			max=java.lang.Math.max(max,((String)fu[i]).length());
		}
		
		for(int i=0;i<co.length;i++){
			max=java.lang.Math.max(max,((String)co[i]).length());
		}
		
		return max;
	}

}//fin de Parseador